;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;; See the file LICENCE for licence information.

;; The cl-store backend. 
(in-package :cl-store)

(defbackend cl-store :magic-number 1953713219
            :stream-type '(unsigned-byte 8)
            :compatible-magic-numbers (1349740876 1414745155)
            :old-magic-numbers (1912923 1886611788 1347635532 1886611820 
                                        1884506444 1347643724 1349732684)
            :extends (resolving-backend)
            :fields ((restorers :accessor restorers 
                                :initform (make-hash-table :size 100))))

(defun register-code (code name &optional (errorp t))
  (aif (and (gethash code (restorers (find-backend 'cl-store))) errorp)
       (error "Code ~A is already defined for ~A." code name)
       (setf (gethash code (restorers (find-backend 'cl-store)))
             name))
  code)

;;  Type code constants
(defvar +referrer-code+ (register-code 1 'referrer nil))
(defvar +unicode-string-code+ (register-code 3 'unicode-string nil))
(defvar +integer-code+ (register-code 4 'integer nil))
(defvar +simple-string-code+ (register-code 5 'simple-string nil))
(defvar +float-code+ (register-code 6 'float nil))
(defvar +ratio-code+ (register-code 7 'ratio nil))
(defvar +character-code+ (register-code 8 'character nil))
(defvar +complex-code+ (register-code 9 'complex nil))
(defvar +symbol-code+ (register-code 10 'symbol nil))
(defvar +cons-code+ (register-code 11 'cons nil))
(defvar +pathname-code+ (register-code 12 'pathname nil))
(defvar +hash-table-code+ (register-code 13 'hash-table nil))
(defvar +standard-object-code+ (register-code 14 'standard-object nil))
(defvar +condition-code+ (register-code 15 'condition nil))
(defvar +structure-object-code+ (register-code 16 'structure-object nil))
(defvar +standard-class-code+ (register-code 17 'standard-class nil))
(defvar +built-in-class-code+ (register-code 18 'built-in-class nil))
(defvar +array-code+ (register-code 19 'array nil))
(defvar +simple-vector-code+ (register-code 20 'simple-vector nil))
(defvar +package-code+ (register-code 21 'package nil))

(defvar +positive-infinity-code+ (register-code 22 'positive-infinity nil))
(defvar +negative-infinity-code+ (register-code 23 'negative-infinity nil))

;; new storing for 32 bit ints
(defvar +32-bit-integer-code+ (register-code 24 '32-bit-integer nil))

(defvar +float-nan-code+ (register-code 25 'nan-float nil))

(defvar +function-code+ (register-code 26 'function nil))
(defvar +gf-code+ (register-code 27 'generic-function nil))

;; Used by SBCL and CMUCL.
(defvar +structure-class-code+ (register-code 28 'structure-class nil))
(defvar +struct-def-code+ (register-code 29 'struct-def nil))

(defvar +gensym-code+ (register-code 30 'gensym nil))

(defvar +positive-double-infinity-code+ (register-code 31 'positive-double-infinity nil))
(defvar +negative-double-infinity-code+ (register-code 32 'negative-double-infinity nil))
(defvar +float-double-nan-code+ (register-code 33 'float-double-nan nil))
(defvar +unicode-base-string-code+ (register-code 34 'unicode-base-string nil))
(defvar +simple-base-string-code+ (register-code 35 'simple-base-string nil))
(defvar +proper-list-code+ (register-code 36 'proper-list))
(defvar +circular-list-code+  (register-code 37 'circular-list))
(defvar +dotted-list-code+  (register-code 38 'dotted-list))



;; setups for type code mapping
(defun output-type-code (code stream)
  (declare (type ub32 code))
  (write-byte (ldb (byte 8 0) code) stream))

(declaim (inline read-type-code))
(defun read-type-code (stream)
  (read-byte stream))

(defmethod referrerp ((backend cl-store) (reader t))
  (eql reader 'referrer))

(defvar *restorers* (restorers (find-backend 'cl-store)))

;; get-next-reader needs to return a symbol which will be used by the
;; backend to lookup the function that was defined by
;; defrestore-cl-store to restore it, or nil if not found. 
(defun lookup-code (code)
  (gethash code *restorers*))

(defmethod get-next-reader ((backend cl-store) (stream stream))
  (declare (optimize speed))
  (let ((type-code (read-type-code stream)))
    (or (lookup-code type-code)
        (error "Type code ~A is not registered." type-code))))


;; referrer, Required for a resolving backend
(defmethod store-referrer ((backend cl-store) (ref t) (stream t))
  (output-type-code +referrer-code+ stream)
  (dump-int ref stream))

(defrestore-cl-store (referrer stream)
  (make-referrer :val (undump-int stream)))


;; integers
;; The theory is that most numbers will fit in 32 bits 
;; so we we have a little optimization for it

;; We need this for circularity stuff.
(defmethod int-or-char-p ((backend cl-store) (type symbol))
  (find type '(integer character 32-bit-integer)))

(defstore-cl-store (obj integer stream)
  (declare (optimize speed (safety 1) (debug 0)))
  (if (typep obj 'sb32)
      (store-32-bit-integer obj stream)
      (store-arbitrary-integer obj stream)))

(defun dump-int (obj stream)
  (declare (optimize speed (safety 0) (debug 0)))
  (typecase obj 
    ((unsigned-byte 8) (write-byte 1 stream) (write-byte obj stream))
    (t (write-byte 2 stream) (store-32-bit obj stream))))

(defun undump-int (stream)
  (declare (optimize speed (safety 0) (debug 0)))
  (ecase (read-byte stream)
    (1 (read-byte stream))
    (2 (read-32-bit stream nil))))

(defun store-32-bit-integer (obj stream)
  (declare (optimize speed (safety 1) (debug 0)) (type sb32 obj))
  (output-type-code +32-bit-integer-code+ stream)
  (write-byte (if (minusp obj) 1 0) stream)
  (dump-int (abs obj) stream))

(defrestore-cl-store (32-bit-integer stream)
  (declare (optimize speed (safety 1) (debug 0)))
  (funcall (if (zerop (the fixnum (read-byte stream))) #'+ #'-)
           (undump-int stream)))

(defun store-arbitrary-integer (obj stream)
  (declare (type integer obj) (stream stream)
           (optimize speed))
  (output-type-code +integer-code+ stream)
  (loop for n = (abs obj) then (ash n -32)
        for counter from 0
        with collect = nil
        until (zerop n)
        do (push n collect)
        finally (progn
                  (store-object (if (minusp obj)
                                    (- counter)
                                    counter)
                                stream)
                  (dolist (num collect)
                    (dump-int num stream)))))

(defrestore-cl-store (integer buff)
  (declare (optimize speed))
  (let ((count (restore-object buff))
        (result 0))
    (declare (type integer result count))
    (loop repeat (abs count) do
          (setf result (the integer (+ (ash result 32) 
                                       (the ub32 (undump-int buff))))))
    (if (minusp count)
        (- result)
        result)))

;; Floats (*special-floats* are setup in the custom.lisp files)
(defvar *special-floats* nil)

(defstore-cl-store (obj float stream)
  (declare (optimize speed))
  (block body
    (let (significand exponent sign)
      (handler-bind ((simple-error
                      #'(lambda (err)
                          (declare (ignore err))
                          (when-let (type (cdr (assoc obj *special-floats*)))
                            (output-type-code type stream)
                            (return-from body)))))
        (multiple-value-setq (significand exponent sign)
          (integer-decode-float obj))
        (output-type-code +float-code+ stream)
        (write-byte (float-type obj) stream)
        (store-object significand stream)
        (store-object (float-radix obj) stream)
        (store-object exponent stream)
        (store-object sign stream)))))

(defrestore-cl-store (float stream)
  (float (* (the float (get-float-type (read-byte stream)))
            (* (the integer (restore-object stream))
               (expt (the integer (restore-object stream))
                     (the integer (restore-object stream))))
            (the integer (restore-object stream)))))

(defun handle-special-float (code name)
  (aif (rassoc code *special-floats*)
       (car it)
       (store-error "~A Cannot be represented." name)))

(defrestore-cl-store (negative-infinity stream)
  (handle-special-float +negative-infinity-code+
                        "Single Float Negative Infinity"))

(defrestore-cl-store (positive-infinity stream)
  (handle-special-float +positive-infinity-code+
                        "Single Float Positive Infinity"))

(defrestore-cl-store (nan-float stream)
   (handle-special-float +float-nan-code+ "Single Float NaN"))

(defrestore-cl-store (negative-double-infinity stream)
  (handle-special-float +negative-double-infinity-code+
                        "Double Float Negative Infinity"))

(defrestore-cl-store (positive-double-infinity stream)
  (handle-special-float +positive-double-infinity-code+
                        "Double Float Positive Infinity"))

(defrestore-cl-store (float-double-nan stream)
  (handle-special-float +float-double-nan-code+
                        "Double Float NaN"))


;; ratio
(defstore-cl-store (obj ratio stream)
  (output-type-code +ratio-code+ stream)
  (store-object (numerator obj) stream)
  (store-object (denominator obj) stream))

(defrestore-cl-store (ratio stream)
  (/ (the integer (restore-object stream))
     (the integer (restore-object stream))))

;; chars
(defstore-cl-store (obj character stream)
  (output-type-code +character-code+ stream)    
  (store-object (char-code obj) stream))

(defrestore-cl-store (character stream)
  (code-char (restore-object stream)))

;; complex
(defstore-cl-store (obj complex stream)
  (output-type-code +complex-code+ stream)    
  (store-object (realpart obj) stream)
  (store-object (imagpart obj) stream))

(defrestore-cl-store (complex stream)
  (complex (restore-object stream)
           (restore-object stream)))

;; symbols
(defstore-cl-store (obj symbol stream)
  (declare (optimize speed))
  (cond ((symbol-package obj)
         (output-type-code +symbol-code+ stream)
         (store-object (symbol-name obj) stream)
         (store-object (package-name (symbol-package obj))
                       stream))
        ;; Symbols with no home package 
        (t (output-type-code +gensym-code+ stream)
           (store-object (symbol-name obj) stream))))

(defrestore-cl-store (symbol stream)
  (values (intern (restore-object stream)
                  (restore-object stream))))

(defrestore-cl-store (gensym stream)
  (make-symbol (restore-object stream)))


;; Lists
(defun dump-proper-list (list length stream)
  (output-type-code +proper-list-code+ stream)
  (store-object length stream)
  (dolist (x list)
    (store-object x stream)))




(defun restore-proper-list (stream)
  (let ((fixes ()))
    (let ((ret (loop for count below (restore-object stream)
                     for elt = (restore-object stream)
                     if (and *check-for-circs* (referrer-p elt))
                       do (push (cons count elt) fixes)
                     collect elt)))
      ;; This requires a bit of fiddling
      (when *check-for-circs*
        (dolist (referrer fixes)
          (let ((ref (cdr referrer))
                (pos (car referrer)))
            (push (delay (setf (nth pos ret)
                               (referred-value ref *restored-values*)))
                  *need-to-fix*))))
      ret)))

(defun dump-dotted-list (list stream)
  (output-type-code +dotted-list-code+ stream)
  (store-object (count-conses list) stream)
  (labels ((rec (list)
             (cond ((atom (cdr list)) ;; last cons cell
                    (store-object (car list) stream)
                    (store-object (cdr list) stream))
                   (t (store-object (car list) stream)
                      (rec (cdr list))))))
    (rec list)))

(defun restore-dotted-list (stream)
  (let* ((ret ())
         (tail ret)
         (conses (restore-object stream)))
    (dotimes (x conses)
      (let ((obj (restore-object stream)))
        (when (and *check-for-circs* (referrer-p obj))
          (let ((x x))
            (push (delay (setf (nth x ret)
                               (referred-value obj *restored-values*)))
                  *need-to-fix*)))
        (if ret
            (setf (cdr tail) (list obj) 
                  tail (cdr tail))
            (setf ret (list obj)
                  tail (last ret)))))
    (setf (cdr tail) (restore-object stream))
    ret))

(defun dump-circular-list (list stream)
  (output-type-code +circular-list-code+ stream)
  (store-object (car list) stream)
  (store-object (cdr list) stream))

(defstore-cl-store (list cons stream)
  (multiple-value-bind (length errorp)
      (proper-list-length list)
    (cond (errorp (dump-dotted-list list stream))
          (length (dump-proper-list list length stream))
          (t (dump-circular-list list stream)))))

(defrestore-cl-store (proper-list stream)
  (restore-proper-list stream))

(defrestore-cl-store (dotted-list stream)
  (restore-dotted-list stream))

(defrestore-cl-store (circular-list stream)
  (resolving-object (ret (cons nil nil))
    (setting (car ret) (restore-object stream))
    (setting (cdr ret) (restore-object stream))))

;; kept for backwards compatibility
(defrestore-cl-store (cons stream)
  (resolving-object (ret (cons nil nil))
    (setting (car ret) (restore-object stream))
    (setting (cdr ret) (restore-object stream))))


;; pathnames
(defstore-cl-store (obj pathname stream)
  (output-type-code +pathname-code+ stream)
  (store-object (pathname-device obj) stream)
  (store-object (pathname-directory obj) stream)
  (store-object (pathname-name obj) stream)
  (store-object (pathname-type obj) stream)
  (store-object (pathname-version obj) stream))

(defrestore-cl-store (pathname stream)
  (make-pathname    
   :device (restore-object stream)
   :directory (restore-object stream)
   :name (restore-object stream)
   :type (restore-object stream)
   :version (restore-object stream)))


;; hash tables
(defstore-cl-store (obj hash-table stream)
  (declare (optimize speed))
  (output-type-code +hash-table-code+ stream)    
  (store-object (hash-table-rehash-size obj) stream)
  (store-object (hash-table-rehash-threshold obj) stream)
  (store-object (hash-table-size obj) stream)
  (store-object (hash-table-test obj) stream)
  (store-object (hash-table-count obj) stream)
  (loop for key being the hash-keys of obj
        using (hash-value value) do
        (store-object key stream)
        (store-object value stream)))

(defrestore-cl-store (hash-table stream)
  (let ((rehash-size (restore-object stream))
        (rehash-threshold (restore-object stream))
        (size (restore-object stream))
        (test (restore-object stream))
        (count (restore-object stream)))
    (declare (type integer count size))
    (let ((hash (make-hash-table :test test
                                 :rehash-size rehash-size
                                 :rehash-threshold rehash-threshold
                                 :size size)))
      (resolving-object (x hash)
        (loop repeat count do
              ;; Unfortunately we can't use the normal setting here
              ;; since there could be a circularity in the key
              ;; and we need to make sure that both objects are 
              ;; removed from the stream at this point.
              (setting-hash (restore-object stream) 
                            (restore-object stream))))
      hash)))

;; Object and Conditions
(defun store-type-object (obj stream)
  (declare (optimize speed))
  (let* ((all-slots (remove-if-not (lambda (x)
                                     (slot-boundp obj (slot-definition-name x)))
                                   (serializable-slots obj)))
         (slots (if *store-class-slots*
                    all-slots
                    (delete-if #'(lambda (x) (eql (slot-definition-allocation x)
                                                  :class))
                               all-slots))))
    (declare (type list slots))
    (store-object (type-of obj) stream)
    (store-object (length slots) stream)
    (dolist (slot slots)
      (let ((slot-name (slot-definition-name slot)))
        (store-object slot-name stream)
        (store-object (slot-value obj slot-name) stream)))))


(defstore-cl-store (obj standard-object stream)
  (output-type-code +standard-object-code+ stream)    
  (store-type-object obj stream))

(defstore-cl-store (obj condition stream)
  (output-type-code +condition-code+ stream)    
  (store-type-object obj stream))

(defun restore-type-object (stream)
  (declare (optimize speed))
  (let* ((class (find-class (restore-object stream)))
         (length (restore-object stream))
         (new-instance (allocate-instance class)))
    (declare (type integer length))
    (loop repeat length do
          (let ((slot-name (restore-object stream)))
            ;; slot-names are always symbols so we don't
            ;; have to worry about circularities
            (resolving-object (obj new-instance)
              (setting (slot-value obj slot-name) (restore-object stream)))))
    new-instance))

(defrestore-cl-store (standard-object stream)
  (restore-type-object stream))

(defrestore-cl-store (condition stream)
  (restore-type-object stream))


;; classes
(defstore-cl-store (obj standard-class stream)
  (output-type-code +standard-class-code+ stream)
  (store-object (class-name obj) stream)
  (store-object (mapcar #'get-slot-details (class-direct-slots obj))
                stream)
  (store-object (mapcar (if *store-class-superclasses*
                            #'identity 
                            #'class-name)
                        (class-direct-superclasses obj))
                stream)
  (store-object (type-of obj) stream))

(defrestore-cl-store (standard-class stream)
  (let* ((class (restore-object stream))
         (slots (restore-object stream))
         (supers (restore-object stream))
         (meta (restore-object stream))
         (keywords '(:direct-slots :direct-superclasses
                     :metaclass))
         (final (mappend #'list keywords (list slots 
                                               (or supers (list 'standard-object))
                                               meta))))
    (cond ((find-class class nil)
           (cond (*nuke-existing-classes*
                  (apply #'ensure-class class final)
                  #+(and clisp (not mop)) (add-methods-for-class class slots))
                 (t (find-class class))))
          (t (apply #'ensure-class class final)
             #+(and clisp (not mop)) (add-methods-for-class class slots)))))

;; built in classes

(defstore-cl-store (obj built-in-class stream)
  (output-type-code +built-in-class-code+ stream)
  (store-object (class-name obj) stream))

#-ecl ;; for some reason this doesn't work with ecl
(defmethod internal-store-object ((backend cl-store) (obj (eql (find-class 'hash-table))) stream)
  (output-type-code +built-in-class-code+ stream)
  (store-object 'cl:hash-table stream))

(defrestore-cl-store (built-in-class stream)
  (find-class (restore-object stream)))



;; Arrays, vectors and strings.
(defstore-cl-store (obj array stream)
  (declare (optimize speed (safety 1) (debug 0)))
  (typecase obj
    (simple-base-string (store-simple-base-string obj stream))
    (simple-string (store-simple-string obj stream))
    (simple-vector (store-simple-vector obj stream))
    (t (store-array obj stream))))

(defun store-array (obj stream)
  (declare (optimize speed (safety 1) (debug 0)))
  (output-type-code +array-code+ stream)
  (if (and (= (array-rank obj) 1)
           (array-has-fill-pointer-p obj))
      (store-object (fill-pointer obj) stream)
      (store-object nil stream))
  (store-object (array-element-type obj) stream)
  (store-object (adjustable-array-p obj) stream)
  (store-object (array-dimensions obj) stream)
  (dolist (x (multiple-value-list (array-displacement obj)))
    (store-object x stream))
  (store-object (array-total-size obj) stream)
  (loop for x from 0 below (array-total-size obj) do
        (store-object (row-major-aref obj x) stream)))

(defrestore-cl-store (array stream)
  (declare (optimize speed (safety 1) (debug 0)))
  (let* ((fill-pointer (restore-object stream))
         (element-type (restore-object stream))
         (adjustable (restore-object stream))
         (dimensions (restore-object stream))
         (displaced-to (restore-object stream))
         (displaced-offset (restore-object stream))
         (size (restore-object stream))
         (res (make-array dimensions  
                          :element-type element-type
                          :adjustable adjustable
                          :fill-pointer fill-pointer)))
    (declare (type cons dimensions) (type array-tot-size size))
    (when displaced-to 
      (adjust-array res dimensions :displaced-to displaced-to
                    :displaced-index-offset displaced-offset))
    (resolving-object (obj res)
      (loop for x from 0 below size do
            (let ((pos x))
              (setting (row-major-aref obj pos) (restore-object stream)))))))

(defun store-simple-vector (obj stream)
  (declare (optimize speed (safety 1) (debug 0))
           (type simple-vector obj))
  (output-type-code +simple-vector-code+ stream)
  (store-object (length obj) stream)
  (loop for x across obj do
    (store-object x stream)))

(defrestore-cl-store (simple-vector stream)
  (declare (optimize speed (safety 1) (debug 0)))
  (let* ((size (restore-object stream))
         (res (make-array size)))
    (declare (type array-size size))
    (resolving-object (obj res)
      (dotimes (i size)
        ;; we need to copy the index so that
        ;; it's value at this time is preserved.
        (let ((x i)) 
          (setting (aref obj x) (restore-object stream)))))
    res))

;; Dumping (unsigned-byte 32) for each character seems
;; like a bit much when most of them will be 
;; base-chars. So we try to cater for them.
(defvar *char-marker* (code-char 255)
  "Largest character that can be represented in 8 bits")

(defun unicode-string-p (string)
  "An implementation specific test for a unicode string."
  (declare (optimize speed (safety 0) (debug 0))
           (type simple-string string))
  #+cmu nil ;; cmucl doesn't support unicode yet.
  #+lispworks (not (typep string 'lw:8-bit-string))
  #-(or cmu lispworks) (some #'(lambda (x) (char> x *char-marker*)) string))

(defun store-simple-string (obj stream)
  (declare (type simple-string obj)
           (optimize speed (safety 1) (debug 0)))
  (cond ((unicode-string-p obj)
         (output-type-code +unicode-string-code+ stream)
         (dump-string #'dump-int obj stream))
        (t (output-type-code +simple-string-code+ stream)
           (dump-string #'write-byte obj stream))))

(defun store-simple-base-string (obj stream)
  (declare (type simple-string obj)
           (optimize speed (safety 1) (debug 0)))
  (cond ((unicode-string-p obj)
         (output-type-code +unicode-base-string-code+ stream)
         (dump-string #'dump-int obj stream))
        (t (output-type-code +simple-base-string-code+ stream)
           (dump-string #'write-byte obj stream))))

(defun dump-string (dumper obj stream)
  (declare (simple-string obj) (function dumper) (stream stream)
           (optimize speed (safety 1) (debug 0)))
  (dump-int (the array-size (length obj)) stream)
  (loop for x across obj do (funcall dumper (char-code x) stream)))

(defrestore-cl-store (simple-string stream)
  (declare (optimize speed))
  (undump-string #'read-byte 'character stream))

(defrestore-cl-store (unicode-string stream)
  (declare (optimize speed))
  (undump-string #'undump-int 'character stream))

(defrestore-cl-store (simple-base-string stream)
  (declare (optimize speed))
  (undump-string #'read-byte 'base-char stream))

(defrestore-cl-store (unicode-base-string stream)
  (declare (optimize speed))
  (undump-string #'undump-int 'base-char stream))

(defun undump-string (reader type stream)
  (declare (type function reader) (type stream stream)
           (optimize speed (safety 1) (debug 0)))
  (let* ((length (the array-size (undump-int stream)) )
         (res (make-string length :element-type type)))
    (declare (type simple-string res))
    (dotimes (x length)
      (setf (schar res x) (code-char (funcall reader stream))))
    res))

;; packages (from Thomas Stenhaug)
(defstore-cl-store (obj package stream)
  (output-type-code +package-code+ stream)  
  (store-object (package-name obj) stream)
  (store-object (package-nicknames obj) stream)
  (store-object (mapcar (if *store-used-packages* #'identity #'package-name)
                        (package-use-list obj))
                stream)
  (store-object (internal-symbols obj) stream)
  (store-object (package-shadowing-symbols obj) stream)
  (store-object (external-symbols obj) stream))

(defun remove-remaining (times stream)
  (declare (optimize speed) (type fixnum times))
  (dotimes (x times)
    (restore-object stream)))

(defrestore-cl-store (package stream)
  (let* ((package-name (restore-object stream))
         (existing-package (find-package package-name)))
    (cond ((or (not existing-package)
               (and existing-package *nuke-existing-packages*))
           (restore-package package-name stream :force *nuke-existing-packages*))
          (t (remove-remaining 5 stream)
             existing-package))))

(defun internal-symbols (package)
  (let ((acc (make-array 100 :adjustable t :fill-pointer 0))
        (used (package-use-list package)))
    (do-symbols (symbol package)
      (unless (find (symbol-package symbol) used)
        (vector-push-extend symbol acc)))
    acc))

(defun external-symbols (package)
  (let ((acc (make-array 100 :adjustable t :fill-pointer 0)))
    (do-external-symbols (symbol package)
      (vector-push-extend symbol acc))
    acc))

(defun restore-package (package-name stream &key force)
  (when (and force (find-package package-name))
    (delete-package package-name))
  (let ((package (make-package package-name
			       :nicknames (restore-object stream)
			       :use (restore-object stream))))
    (loop for symbol across (restore-object stream) do
      (import symbol package))
    (shadow (restore-object stream) package)
    (loop for symbol across (restore-object stream) do
      (export symbol package))
    package))

;; Function storing hack.
;; This just stores the function name if we can find it
;; or signal a store-error.
(defun parse-name (name)
  (let ((name (subseq name 21)))
    (declare (type simple-string name))
    (if (search name "SB!" :end1 3)
        (replace name "SB-" :end1 3)
        name)))

#+sbcl
(defvar *sbcl-readtable* (copy-readtable *readtable*))
#+sbcl
(set-macro-character #\# #'(lambda (c s) 
                             (declare (ignore c s))
                             (store-error "Invalid character in function name."))
                     nil
                     *sbcl-readtable*)

(defun get-function-name (obj)
  (multiple-value-bind (l cp name) (function-lambda-expression obj) 
    (declare (ignore l cp))
    (cond ((and name (or (symbolp name) (consp name))) name)
          ;;  Try to deal with sbcl's naming convention
          ;; of built in functions (pre 0.9)
          #+sbcl
          ((and name (stringp name)
                (search "top level local call " (the simple-string name)))
           (let ((new-name (parse-name name))
                 (*readtable* *sbcl-readtable*))
             (unless (string= new-name "")
               (handler-case (read-from-string new-name)
                 (error (c) (declare (ignore c))
                        (store-error "Unable to determine function name for ~A."
                                     obj))))))
          (t (store-error "Unable to determine function name for ~A."
                          obj)))))
  

(defstore-cl-store (obj function stream)
  (output-type-code +function-code+ stream)
  (store-object (get-function-name obj) stream))



(defrestore-cl-store (function stream)
  (fdefinition (restore-object stream)))

;; Generic function, just dumps the gf-name
(defstore-cl-store (obj generic-function stream)
  (output-type-code +gf-code+ stream)
  (aif (generic-function-name obj)
       (store-object it stream)
       (store-error "No generic function name for ~A." obj)))

(defrestore-cl-store (generic-function stream)
  (fdefinition (restore-object stream)))


(setf *default-backend* (find-backend 'cl-store))

;; EOF
