;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;; See the file LICENCE for licence information.

(in-package :cl-store)

; special floats
(defvar +single-positive-infinity+ most-positive-single-float)
(defvar +single-negative-infinity+ most-negative-single-float)
(defvar +single-nan+)

(defvar +double-positive-infinity+ most-positive-double-float)
(defvar +double-negative-infinity+ most-negative-double-float)
(defvar +double-nan+)

(sb-int:with-float-traps-masked (:overflow :invalid)
  (setf +single-positive-infinity+ (expt +single-positive-infinity+ 2))
  (setf +single-negative-infinity+ (expt +single-negative-infinity+ 3))
  (setf +single-nan+ (/ +single-negative-infinity+ +single-negative-infinity+))
  (setf +double-positive-infinity+ (expt +double-positive-infinity+ 2))
  (setf +double-negative-infinity+ (expt +double-negative-infinity+ 3))
  (setf +double-nan+ (/ +double-negative-infinity+ +double-negative-infinity+)))

(setf *special-floats*
  (list (cons +double-positive-infinity+ +positive-double-infinity-code+)
        (cons +single-positive-infinity+ +positive-infinity-code+)
        (cons +single-negative-infinity+ +negative-infinity-code+)
        (cons +double-negative-infinity+ +negative-double-infinity-code+)
        (cons +single-nan+ +float-nan-code+)
        (cons +double-nan+ +float-double-nan-code+)))



;; Custom structure storing

(defstore-cl-store (obj structure-object stream)
  (output-type-code +structure-object-code+ stream)
  (store-type-object obj stream))

(defrestore-cl-store (structure-object stream)
  (restore-type-object stream))


;; Structure definition storing
(defun get-layout (obj)
  (slot-value obj 'sb-pcl::wrapper))

(defun get-info (obj)
  (declare (type sb-kernel:layout obj))
  (slot-value obj 'sb-int:info))

(defun dd-name (dd)
  (slot-value dd 'sb-kernel::name))

(defvar *sbcl-struct-inherits*
  (list (get-layout (find-class t))
        (get-layout (find-class 'sb-kernel:instance))
        (get-layout (find-class 'cl:structure-object))))

(defstruct (struct-def (:conc-name sdef-))
  (supers (required-arg :supers) :type list)
  (info (required-arg :info) :type sb-kernel:defstruct-description))

(defun info-or-die (obj)
  (let ((wrapper (get-layout obj)))
    (if wrapper
        (or (get-info wrapper) 
            (store-error "No defstruct-definition for ~A." obj))
        (store-error "No wrapper for ~A." obj))))

(defun save-able-supers (obj)
  (set-difference (coerce (slot-value (get-layout obj) 'sb-kernel::inherits)
                          'list)
                  *sbcl-struct-inherits*))

(defun get-supers (obj)
  (loop for x in (save-able-supers obj) 
     collect (let ((name (dd-name (get-info x))))
               (if *store-class-superclasses* 
                   (find-class name)
                   name))))

(defstore-cl-store (obj structure-class stream)
  (output-type-code +structure-class-code+ stream)
  (store-object (make-struct-def :info (info-or-die obj)
                                 :supers (get-supers obj))
                stream))

(defstore-cl-store (obj struct-def stream)
  (output-type-code +struct-def-code+ stream)
  (store-object (sdef-supers obj) stream)
  (store-object (sdef-info obj) stream))

;; Restoring 
(defun sbcl-struct-defs (info)
  (append (sb-kernel::constructor-definitions info)
          (sb-kernel::class-method-definitions info)))

(defun create-make-foo (dd)
  (declare (optimize speed))
  (funcall (compile nil `(lambda () ,@(sbcl-struct-defs dd))))
  (find-class (dd-name dd)))

(defun sbcl-define-structure (dd supers)
  (cond ((or *nuke-existing-classes*  
             (not (find-class (dd-name dd) nil)))
         ;; create-struct
         (sb-kernel::%defstruct dd supers)
         ;; compiler stuff
         (sb-kernel::%compiler-defstruct dd supers) 
         ;; create make-?
         (create-make-foo dd))
        (t (find-class (dd-name dd)))))
         
(defun super-layout (super)
  (etypecase super
    (symbol (get-layout (find-class super)))
    (structure-class 
     (super-layout (dd-name (info-or-die super))))))

(defun super-layouts (supers)
  (loop for super in supers 
     collect (super-layout super)))

(defrestore-cl-store (structure-class stream)
  (restore-object stream))
    
(defrestore-cl-store (struct-def stream)
  (let* ((supers (super-layouts (restore-object stream)))
         (dd (restore-object stream)))
    (sbcl-define-structure dd (if supers 
                                  (coerce (append  *sbcl-struct-inherits*
                                                   supers)
                                          'vector)
                                  (coerce *sbcl-struct-inherits* 'vector)))))

;; EOF